/*
FRAM byte read/write demo
=====================================
 FRAM support routines:
 1 - void fram_read_bytes(spi_inst_t *spi, uint cs_pin, uint32_t addr, uint8_t *buf, uint8_t len)
 ---- reads len bytes from array buf starting at addr
 
 2- void fram_write_bytes(spi_inst_t *spi, uint cs_pin, uint32_t addr, uint8_t * data, uint8_t len)
  ---- writes len bytes from array data starting at addr

3- void fram_write_enable(spi_inst_t *spi, uint cs_pin)
4- void fram_write_disable(spi_inst_t *spi, uint cs_pin)
---- enables/disables the write operation above
=====================================
 threads on core0:
 1- the usual blinky and prints if there are lock errors
 2- FRAM test code
 ---- sends three counts to FRAM locations 0-2
 ---- reads them back and prints them
 ---- data length is one byte, the values wrap after 255
 ====================================
Connections:
gpio 0 --> default uart transmit - dongle white wire
gpoi 1 --> default uart areceife - dongle green wire

gpio 10 spi1 sck --> FRAM SCK
gpio 11 spi1 sdo --> FRAM MOSI
gpio 12 spi1 sdi --> FRAM MISO
gpio 13 chip sel --> FRAM CS

FRAM is MB85RS4MT on ADAFRUIT 4719 breakout
https://cdn-shop.adafruit.com/product-files/4719/4719_MB85RS4MT.pdf

VBUS on PICO --> FRAM Vin  (5 volt)
FRAM WP connected to its 3v3 pin
FRAM HOLD connected to its 3v3 pin

The spi channel appears to run correctly at 20MHz on a white board.
So: 32 clocks to set up a read/write, then 8 clocks/byte read/write
This imples a max read/write rate of 2.5 MByte/sec for larger arrays
and 0.5 Mbyte/sec for single bytes.
(Probably runs at 30 MHz, but fails at 40 MHz. The FRAM is rated for
40 MHz, but the white board is not)

 */

#include "hardware/gpio.h"
#include "hardware/timer.h"

#include "pico/stdlib.h"
#include "hardware/uart.h"
#include "hardware/spi.h"
#include "stdio.h"
#include <string.h>
#include <pico/multicore.h>
#include "hardware/sync.h"
#include "stdlib.h"

// ==========================================
// === protothreads globals
// ==========================================
// protothreads header
#include "pt_cornell_rp2040_v1_1_1.h"

// ==================================================
// === user's serial input thread
// ==================================================
// serial_read an serial_write do not block any thread
// except this one
static PT_THREAD (protothread_serial(struct pt *pt))
{
    PT_BEGIN(pt);
      static int test_in1, test_in2, sum ;
      //
      // start the serial i/o

      //
      while(1) {
        // print prompt
        sprintf(pt_serial_out_buffer, "input two numbers: ");
        // spawn a thread to do the non-blocking write
        serial_write ;

        // spawn a thread to do the non-blocking serial read
         serial_read ;
        // convert input string to number
        sscanf(pt_serial_in_buffer,"%d %d", &test_in1, &test_in2) ;

        // add and convert back to sring
        sprintf(pt_serial_out_buffer,"sum = %d\r\n", test_in1 + test_in2);
        // spawn a thread to do the non-blocking serial write
         serial_write ;


        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // timer thread

// ==================================================
// === toggle25 thread 
// ==================================================
// the on-board LED blinks
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
    PT_BEGIN(pt);
    static bool LED_state = false ;
    
     // set up LED p25 to blink
     gpio_init(25) ;	
     gpio_set_dir(25, GPIO_OUT) ;
     gpio_put(25, true);
     // data structure for interval timer
     PT_INTERVAL_INIT() ;

      while(1) {
        // yield time 0.1 second
        //PT_YIELD_usec(100000) ;
        PT_YIELD_INTERVAL(100000) ;

        // toggle the LED on PICO
        LED_state = LED_state? false : true ;
        gpio_put(25, LED_state);
        //
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // blink thread

// ==================================================
// === spi FRAM support 
// ==================================================
//SPI configurations
#define PIN_SCK  10
#define PIN_SDO 11
#define PIN_SDIN 12
#define PIN_CS   13
#define SPI_PORT spi1

// chip select macros
#define CS_on  gpio_put(PIN_CS,0) 
#define CS_off gpio_put(PIN_CS,1) 

// FRAM commands
#define fram_write (2)
#define fram_read  (3)
#define fram_we    (6)
#define fram_wd    (4)

void fram_read_bytes(spi_inst_t *spi, uint cs_pin, uint32_t addr, uint8_t *buf, uint8_t len) {
    CS_on ;
    uint8_t cmdbuf[4] = {
            fram_read,
            addr >> 16,
            addr >> 8,
            addr
    };
    spi_write_blocking(spi, cmdbuf, 4);
    spi_read_blocking(spi, 0, buf, len);
    CS_off ; 
}

void fram_write_enable(spi_inst_t *spi, uint cs_pin) {
    CS_on ; 
    uint8_t cmd = fram_we ;
    spi_write_blocking(spi, &cmd, 1);
    CS_off ; 
}

void fram_write_disable(spi_inst_t *spi, uint cs_pin) {
    CS_on ; 
    uint8_t cmd = fram_wd ;
    spi_write_blocking(spi, &cmd, 1);
    CS_off ; 
}

void fram_write_bytes(spi_inst_t *spi, uint cs_pin, uint32_t addr, uint8_t * data, uint8_t len) {
    uint8_t cmdbuf[4] = {
            fram_write,
            addr >> 16,
            addr >> 8,
            addr
    };
    fram_write_enable(SPI_PORT, cs_pin) ;
    CS_on ;
    spi_write_blocking(spi, cmdbuf, 4) ;
    spi_write_blocking(spi, data, len) ;
    CS_off ;
    fram_write_disable(SPI_PORT, cs_pin) ;
}

// ==========================================
// === FRAM read/write example thread
// ==========================================
// SPI attached FRAM memory
static PT_THREAD (protothread_fram(struct pt *pt))
{
    PT_BEGIN(pt);
    // generate sequential numbers for testing
    static int write_count; 

    // address is 19 bits
    static int fram_address ;
    // buffers for read/write
    static uint8_t fram_data[256] ;
    static uint8_t read_buffer[256] ;

    // Initialize SPI channel (channel, baud rate set to 20MHz)
    spi_init(SPI_PORT, 20000000) ;

    // Map SPI signals to GPIO ports
    //gpio_set_function(PIN_MISO, GPIO_FUNC_SPI);
    gpio_set_function(PIN_SCK, GPIO_FUNC_SPI);
    gpio_set_function(PIN_SDO, GPIO_FUNC_SPI);
    gpio_set_function(PIN_SDIN, GPIO_FUNC_SPI);
    // toggle CS manually rather than use framed CS
    gpio_init(PIN_CS) ;	
    // turn off chip sel
    gpio_put(PIN_CS, true) ;
    gpio_set_dir(PIN_CS, GPIO_OUT) ; 
    

     // data structure for interval timer
     PT_INTERVAL_INIT() ;

      while(1) {
        // yield time
        PT_YIELD_INTERVAL(500000) ;
        write_count++ ; 
        // set base address
        fram_address = 500000 ;
        fram_data[0] = write_count ;
        fram_data[1] = write_count-1 ;
        fram_data[2] = write_count-2 ;
        //
        // Write data to FRAM
        fram_write_bytes(SPI_PORT, PIN_CS, fram_address, fram_data, 3) ;
        // read it back
        fram_read_bytes(SPI_PORT, PIN_CS, fram_address, read_buffer, 3) ;
        // print it
        printf("%d %d %d\n\r", read_buffer[0], read_buffer[1], read_buffer[2]) ;

        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // blink thread


// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  
  //  === add threads  ====================
  // for core 1
  //pt_add_thread(protothread_toggle_fast);
  //pt_add_thread(protothread_toggle3);
  //pt_add_thread(protothread_sum) ;
  //
  // === initalize the scheduler ==========
  //pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock
  //set_sys_clock_khz(250000, true); // 171us
  
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.1 two-core\n\r");
  
  // start core 1 threads
  //multicore_reset_core1();
  //multicore_launch_core1(&core1_main);

  // === config threads ========================
  // for core 0
  //pt_add_thread(protothread_serial);
  pt_add_thread(protothread_toggle25);
  //pt_add_thread(protothread_toggle2);
  pt_add_thread(protothread_fram);
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main
///////////
// end ////
///////////